"""Implementation of Rule L050."""
from typing import Optional

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import RootOnlyCrawler
from sqlfluff.utils.functional import Segments, sp, rsp
from sqlfluff.core.rules.doc_decorators import document_fix_compatible, document_groups


@document_groups
@document_fix_compatible
class Rule_L050(BaseRule):
    """Files must not begin with newlines or whitespace.

    **Anti-pattern**

    The file begins with newlines or whitespace. The ``^``
    represents the beginning of the file.

    .. code-block:: sql
       :force:

        ^

        SELECT
            a
        FROM foo

        -- Beginning on an indented line is also forbidden,
        -- (the • represents space).

        ••••SELECT
        ••••a
        FROM
        ••••foo

    **Best practice**

    Start file on either code or comment. (The ``^`` represents the beginning
    of the file.)

    .. code-block:: sql
       :force:


        ^SELECT
            a
        FROM foo

        -- Including an initial block comment.

        ^/*
        This is a description of my SQL code.
        */
        SELECT
            a
        FROM
            foo

        -- Including an initial inline comment.

        ^--This is a description of my SQL code.
        SELECT
            a
        FROM
            foo
    """

    groups = ("all",)
    targets_templated = True
    # Use the RootOnlyCrawler to only call _eval() ONCE, with the root segment.
    crawl_behaviour = RootOnlyCrawler()
    lint_phase = "post"

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Files must not begin with newlines or whitespace."""
        # Only check raw segments. This ensures we don't try and delete the same
        # whitespace multiple times (i.e. for non-raw segments higher in the
        # tree).
        raw_segments = []
        whitespace_types = {"newline", "whitespace", "indent", "dedent"}
        for seg in context.segment.recursive_crawl_all():
            if not seg.is_raw():
                continue

            if seg.is_type(*whitespace_types):
                raw_segments.append(seg)
                continue

            segment = Segments(seg)
            raw_stack = Segments(*raw_segments, templated_file=context.templated_file)
            # Non-whitespace segment.
            if (
                not raw_stack.all(sp.is_meta())
                # Found leaf of parse tree.
                and not segment.all(sp.is_expandable())
                # It is possible that a template segment (e.g.
                # {{ config(materialized='view') }}) renders to an empty string
                # and as such is omitted from the parsed tree. We therefore
                # should flag if a templated raw slice intersects with the
                # source slices in the raw stack and skip this rule to avoid
                # risking collisions with template objects.
                and not raw_stack.raw_slices.any(rsp.is_slice_type("templated"))
            ):
                return LintResult(
                    anchor=context.segment,
                    fixes=[LintFix.delete(d) for d in raw_stack],
                )
            else:
                break
        return None



"""Tests for the SQLFluff integration with the "diff-quality" tool."""
import sys
from pathlib import Path

import pytest

from sqlfluff import diff_quality_plugin
from sqlfluff.cli.commands import lint
from sqlfluff.utils.testing.cli import invoke_assert_code


@pytest.mark.parametrize(
    "sql_path,expected_violations_lines",
    [
        ("linter/indentation_errors.sql", list(range(2, 7))),
        ("linter/parse_error.sql", {1}),
        # NB: This version of the file is in a directory configured
        # to ignore parsing errors.
        ("linter/diffquality/parse_error.sql", []),
    ],
)
def test_diff_quality_plugin(sql_path, expected_violations_lines, monkeypatch):
    """Test the plugin at least finds errors on the expected lines."""

    def execute(command, exit_codes):
        printable_command_parts = [
            c.decode(sys.getfilesystemencoding()) if isinstance(c, bytes) else c
            for c in command
        ]

        result = invoke_assert_code(
            ret_code=1 if expected_violations_lines else 0,
            args=[
                lint,
                printable_command_parts[2:],
            ],
        )
        return result.output, ""

    # Mock the execute function -- this is an attempt to prevent the CircleCI
    # coverage check from hanging. (We've seen issues in the past where using
    # subprocesses caused things to occasionally hang.)
    monkeypatch.setattr(diff_quality_plugin, "execute", execute)
    monkeypatch.chdir("test/fixtures/")
    violation_reporter = diff_quality_plugin.diff_cover_report_quality(
        options="--processes=1"
    )
    sql_path = str(Path(sql_path))

    violations_dict = violation_reporter.violations_batch([sql_path])
    assert isinstance(violations_dict, dict)
    if expected_violations_lines:
        assert len(violations_dict[sql_path]) > 0
        violations_lines = {v.line for v in violations_dict[sql_path]}
        for expected_line in expected_violations_lines:
            assert expected_line in violations_lines
    else:
        assert len(violations_dict[sql_path]) == 0
